#!/usr/bin/ruby

# update-kernel.rb, v0.1
# 
# Steffen Rusitschka, http://www.ruzee.com, GPL2 license
# 
# This is a Ruby script that automatically compiles, installs (via genkernel)
# the latest kernel found in /usr/src/. Existing .config files are copied
# from the previously used kernel and the symlink gets updated.
# Afterwards, /boot/grub/grub.conf gets updated (please check manually before 
# booting).
# Optionally you my specify a splash image to use and kernel modules that get
# automatically emerged.
# 
# Configuration example file /etc/update-kernel.conf (please remove the 
# first '#' comment character ;-)):
# 
# # your boot partition in grub syntax
# $boot_partition = "(hd0,0)"
# 
# # your root partition in /etc/fstab syntax
# $root_partition = "/dev/hda3"
# 
# # the splash image you want to use (optional)
# $splash_theme = "emergence"
# 
# # additional parameters to pass to genkernel (optional)
# $genkernel_parameters = "--menuconfig"
# 
# # additional kernel parameters to put into grub.conf (optional)
# $kernel_parameters = "vga=0x317 ramdisk=8192"
# 
# # kernel modules to emerge after a kernel update (optional, 
# # known to work with alsa and nvidia)
# $modules = "alsa-driver nvidia-drivers"


require 'fileutils'

MAGIC = "### AUTO-GENERATED BY UPDATE-KERNEL "
MAGIC_BEGIN = MAGIC + "BEGINS"
MAGIC_END = MAGIC + "ENDS"
GRUB_CONF = "/boot/grub/grub.conf"

$boot_partition = "(hd0,0)"
$root_partition = "/dev/hda3"
load "/etc/update-kernel.conf" if File.exist?("/etc/update-kernel.conf")

class KernelVersion
  include Comparable

  attr_accessor :major, :minor, :minor2, :release, :name
  
  def initialize(s)
    @major, @minor, @minor2, @name, temp, @release = 
      s.scan(/^(\d+)\.(\d+)\.(\d+)-(\w+)(-r(\d+))?$/)[0].collect { |s| 
        s =~ /\d+/ ? s.to_i : s
      }
    @release = 0 unless @release
  end
  
  def to_s
    out = "#{@major}.#{@minor}.#{@minor2}-#{@name}"
    out << "-r#{@release}" unless @release == 0
    out
  end
  
  def <=>(other)
    [self.major, self.minor, self.minor2, self.release] <=> 
      [other.major, other.minor, other.minor2, other.release]
  end
end

def detect_arch
  $arch = `uname -m`.chomp
  $arch = "x86" if $arch =~ /^i\d86$/
  $arch = "mips" if $arch =~ /^mips/
  puts "Detected architecture is '#{$arch}'."
end

def detect_kernels
  src_kernel_paths = Dir.entries("/usr/src").delete_if do |path|
    path = "/usr/src/#{path}"
    !FileTest.directory?(path) || FileTest.symlink?(path) || (path =~ /\/\.\.?$/)
  end

  $src_kernels = src_kernel_paths.collect do |kernel|
    kernel =~ /^linux\-/
    KernelVersion.new($')
  end.sort
  
  File.readlink("/usr/src/linux") =~ /linux\-/
  $src_cur_kernel = KernelVersion.new($')
end

def update_symlink
  latest_kernel = $src_kernels[-1]
  puts "Latest installed kernel is #{latest_kernel}."
  if latest_kernel.to_s == $src_cur_kernel.to_s
    $src_active_kernel = $src_cur_kernel
    puts "Using already symlinked kernel version #{$src_cur_kernel}."
  else
    $src_active_kernel = $src_kernels[-1]
    puts "Updating kernel symlink from version #{$src_cur_kernel} to #{$src_active_kernel}"
    unless $test_mode
      File.delete "/usr/src/linux"
      File.symlink "linux-#{$src_active_kernel}", "/usr/src/linux"
      unless File.exist?("/usr/src/linux/.config")
        puts "No kernel .config exists. Copying old kernel .config."
        FileUtils.cp "/usr/src/linux-#{$src_cur_kernel}/.config", "/usr/src/linux/"
      end
    end
  end
end

def gen_kernel
  command = "nice genkernel --oldconfig #{$genkernel_parameters}"
  command << " --gensplash=#{$splash_theme}" if $splash_theme
  command << " all"
  puts "Executing '#{command}'..."
  system "#{command}" unless $test_mode
end

def grub_entry_for_kernel(kernel)
  return "" if kernel.nil?
  out = "title=Gentoo Linux #{kernel}\n"
  out << "root #{$boot_partition}\n"
  out << "kernel /kernel-genkernel-#{$arch}-#{kernel} root=/dev/ram0 init=/linuxrc real_root=#{$root_partition}"
  out <<   " splash=verbose,fadein,theme:#{$splash_theme}" if $splash_theme
  out <<   " #{$kernel_parameters}\n"
  out << "initrd /initramfs-genkernel-#{$arch}-#{kernel}\n\n"
end

def update_grub
  out = ""
  inject = "#{MAGIC_BEGIN}\n\n"
  -1.downto(-3) do |i|
    inject << grub_entry_for_kernel($src_kernels[i])
  end
  inject << "#{MAGIC_END}\n"
  injected = false
  File.open(GRUB_CONF, "r") do |grub_conf|
    while line = grub_conf.gets
      if line =~ /#{MAGIC_BEGIN}/ .. line =~ /#{MAGIC_END}/
        if !injected
          injected = true
          out << inject
        end
      else
        out << line
      end
    end
    out << inject unless injected
  end
  puts "Updating '#{GRUB_CONF}'."
  if $test_mode
    puts inject
  else
    File.rename GRUB_CONF, GRUB_CONF + ".update-kernel-backup"
    File.open(GRUB_CONF, "w+") do |grub_conf|
      grub_conf << out
    end
  end
end

def emerge_modules
  if $modules
    puts "Emerging modules '#{$modules}'."
    system "nice emerge -v #{$modules}" unless $test_mode
  end
end

$test_mode = `whoami`.chomp != "root"
puts "\n*** NOT RUNNING AS 'root'. USING TEST MODE. ***\n\n" if $test_mode
detect_arch
detect_kernels
update_symlink
gen_kernel
update_grub
emerge_modules
puts "\nDone.\n\n"
